/**
 * @file
 * Autocomplete flexible plugin.
 */

/**
 * Cardinality maximum (no limit !).
 *
 * @type {number}
 */
const AF_CARDINALITY_MAX = -1;

/**
 * Default settings:
 *   - autocompleteDisabledValue: Define a value when autocomplete input is disabled.
 *   - buttonRemoveLabel: Remove item button label.
 *   - max: The max values of the form element.
 *   - minLength: The minimum of characters to trigger autocomplete.
 *   - selectedListClass: The selected list wrapper class.
 *
 * @type {object}
 */
const AF_DEFAULTS = {
  max: AF_CARDINALITY_MAX,
  minLength: 3,
  autocompleteDisabledValue: "",
  duplicationErrorMessage: Drupal.t("This item has already been added."),
  buttonRemoveLabel: Drupal.t("Remove"),
  buttonRemoveClass: "autocomplete-flexible-button-delete",
  selectedListClass: "autocomplete-flexible-selected-list",
  wrapperClass: "autocomplete-flexible-wrapper",
};

/**
 * The plugin initialization class.
 * Added to wrapper on plugin initialization.
 *
 * @type {string}
 */
const AF_PLUGIN_INIT_CLASS = "autocomplete-flexible-init";

/**
 * Local object for method references and define script meta-data.
 *
 * @return {object}
 *   The autocomplete flexible plugin.
 */
// eslint-disable-next-line no-unused-vars
class AutocompleteFlexible {

  /**
   * Plugin constructor.
   *
   * @param {jQuery|HTMLElement} $inputAutocomplete
   *  The input autocomplete element.
   * @param {jQuery|HTMLElement} $inputHidden
   *  The input hidden element.
   * @param {object} options
   *   Custom settings (override AF_DEFAULTS).
   * @param {object} values
   *   The available values to append to the bottom of field.
   */
  constructor($inputAutocomplete, $inputHidden, options, values) {
    this._settings = Object.assign(AF_DEFAULTS, options);
    this._values = values;

    this.$inputAutocomplete = $inputAutocomplete;
    this.$inputHidden = $inputHidden;
    this.$wrapper = $inputHidden.parents("." + this._settings.wrapperClass);
    this.$selectedValues = false;

    // Clone Drupal autocomplete.
    this._autocomplete = Object.assign({}, Drupal.autocomplete);
    this._autocomplete.options.minLength = this._settings.minLength;
    this._autocomplete.options.renderItem = this.renderItem;
    this._autocomplete.options.select = this.selectHandler;

    this.message = false;
  }

  /**
   * Append an item.
   *
   * @param {Object|String} item
   *   The selected item.
   *
   * @private
   */
  _appendItem(item) {
    const that = this;

    const $item = jQuery("<li>" +
      Drupal.theme("autocompleteFlexibleSelectedItem", item) +
      "<button class=\"" + this._settings.buttonRemoveClass + "\" data-autocomplete-flexible-value=\"" + item.value + "\">" +
      "<span>" + this._settings.buttonRemoveLabel + "</span>" +
      "</button>" +
      "</li>");

    this.$selectedValues.append($item);

    // Add value in input hidden field.
    this._values[item.value] = item;
    this.$inputHidden.val(Object.keys(this._values).join(","));

    // Disable input autocomplete.
    this._disableInputAutocomplete();

    // Manage item removing.
    jQuery("." + this._settings.buttonRemoveClass, $item).on("click", event => {
      event.preventDefault();

      const value = jQuery(event.target).attr("data-autocomplete-flexible-value");

      that._removeItem(that._values[value]);
    });
  }

  /**
   * Disable input autocomplete.
   *
   * @private
   */
  _disableInputAutocomplete() {
    if (
      this._settings.max !== AF_CARDINALITY_MAX &&
      Object.keys(this._values).length >= this._settings.max
    ) {
      this.$inputAutocomplete.prop("disabled", true);

      // Add a description text.
      this.$inputAutocomplete.val(this._settings.autocompleteDisabledValue);
    }
  }

  /**
   * Display default values.
   *
   * @private
   */
  _displayDefaultValues() {
    const that = this;

    // Init selected list.
    if (!this.$selectedValues) {
      this.$selectedValues = jQuery("<ul class=\"" + this._settings.selectedListClass + "\"></ul>");
      this.$wrapper.append(this.$selectedValues);
    }

    Object.entries(this._values).forEach(([value, item]) => {
      // Prevent "value" entry doesn't exist.
      item.value = value;
      that._values[value].value = value;

      that._appendItem(item);
    });
  }

  /**
   * Enable input autocomplete.
   *
   * @private
   */
  _enableInputAutocomplete() {
    if (
      (
        this._settings.max === AF_CARDINALITY_MAX ||
        Object.keys(this._values).length < this._settings.max
      ) &&
      this.$inputAutocomplete.prop("disabled")
    ) {
      this.$inputAutocomplete.prop("disabled", false);

      // Remove description text.
      this.$inputAutocomplete.val("");
    }
  }

  /**
   * Initialize plugin.
   */
  init() {
    // Make sure to initialize the plugin only once for this element.
    if (
      this.$wrapper.length === 0 ||
      this.$wrapper.hasClass(AF_PLUGIN_INIT_CLASS)
    ) {
      return;
    }
    this.$wrapper.addClass(AF_PLUGIN_INIT_CLASS);

    // Initialize autocomplete management.
    this._initAutocomplete();

    this._initMessage();

    // Display default values.
    this._displayDefaultValues();

    // Make sure autocomplete input is empty.
    this.$inputAutocomplete.val("");

    // Disable input autocomplete if max values is reached.
    this._disableInputAutocomplete();
  }

  /**
   * Initialize autocomplete management.
   *
   * @private
   */
  _initAutocomplete() {
    const that = this;

    // Destroy Drupal Core autocomplete on this field.
    this.$inputAutocomplete.autocomplete("destroy");

    // Manage autocomplete flexible engine.
    this.$inputAutocomplete.autocomplete(this._autocomplete.options).each(function () {
      jQuery(this).data("ui-autocomplete")._renderItem = that._autocomplete.options.renderItem;
    });
    this.$inputAutocomplete.on("compositionstart.autocomplete", () => {
      that._autocomplete.options.isComposing = true;
    });
    this.$inputAutocomplete.on("compositionend.autocomplete", () => {
      that._autocomplete.options.isComposing = false;
    });

    // Manage autocomplete selection.
    this.$inputAutocomplete.on("autocompleteselect", (event, ui) => {
      // Clear old message.
      this.message.clear();

      // Check value is already added.
      if (
        that._settings.duplicationErrorMessage &&
        that._values[ui.item.value] !== undefined
      ) {
        this.message.add(that._settings.duplicationErrorMessage, {
          type: "error"
        });
        return;
      }

      // Append value in display list.
      that._appendItem(ui.item);
    });
  }

  /**
   * Initialize message management.
   *
   * @private
   */
  _initMessage() {
    const messageWrapper = document.createElement("div");
    this.$wrapper.prepend(messageWrapper);

    this.message = new Drupal.Message(messageWrapper);
  }

  /**
   * Remove an item.
   *
   * @param {Object|String} item
   *   The item to remove.
   *
   * @private
   */
  _removeItem(item) {
    // Remove item in input hidden value.
    delete this._values[item.value];
    this.$inputHidden.val(Object.keys(this._values).join(","));

    // Remove item in selected list.
    jQuery("button[data-autocomplete-flexible-value=\"" + item.value + "\"]")
      .parent("li")
      .remove();

    // Enable input autocomplete.
    this._enableInputAutocomplete();
  }

  /**
   * Override jQuery UI _renderItem function to output HTML by default.
   *
   * @param {jQuery} ul
   *   jQuery collection of the ul element.
   * @param {object} item
   *   The list item to append.
   *
   * @return {jQuery}
   *   jQuery collection of the ul element.
   */
  renderItem(ul, item) {
    // Give possibility to customize item result.
    return jQuery("<li>").append(jQuery("<a>").html(
      Drupal.theme("autocompleteFlexibleItemResult", item)
    )).appendTo(ul);
  }

  /**
   * Handles an autocompleteselect event.
   *
   * @param {jQuery.Event} event
   *   The event triggered.
   * @param {object} ui
   *   The jQuery UI settings object.
   *
   * @return {bool}
   *   Returns false to indicate the event status.
   *
   * @see https://api.jqueryui.com/autocomplete/#event-select
   */
  // eslint-disable-next-line no-unused-vars
  selectHandler(event, ui) {
    // Return false to tell jQuery UI that we've filled in the value already.
    return false;
  }

}

/**
 * Render autocomplete flexible item result.
 * It's possible to return any HTML rendering.
 *
 * @param {Object|String} item
 *   An item result.
 *
 * @return {string|HTMLElement}
 *   The item result rendering.
 */
Drupal.theme.autocompleteFlexibleItemResult = (item) => {
  // By default, return the value itself.
  return item.value;
}

/**
 * Render autocomplete flexible item result.
 * It's possible to return any HTML rendering.
 *
 * @param {Object|String} item
 *   The selected item.
 *
 * @return {string|HTMLElement}
 *   The selected value rendering.
 */
Drupal.theme.autocompleteFlexibleSelectedItem = (item) => {
  // By default, return the value itself.
  return item.value;
}
